#include    <stdio.h>
#include    <unistd.h>
#include    <sys/types.h>
#include    <sys/stat.h>
#include    <dirent.h>
#include    <time.h>
#include    <pwd.h>

#define	TRUE	1
#define	FALSE	0
#define	ERR	(-1)

	int	sym_disp = TRUE;
	int	ext_disp = TRUE;
	int	red_disp = TRUE;
	int	uid_disp = TRUE;
	int	sort_flg = TRUE;
	char	*prog = "tree";
	char	nest[256];
	char	cwd[256];

typedef	struct _DIR {
	struct _DIR	*next;
	char		name[1];
} DIRPTR;

DIRPTR	*dir_sort(DIRPTR *np)
{
    DIRPTR    *tp, *bp;
    DIRPTR    *lp, *rp;

    if ( np == NULL || np->next == NULL )
	return np;

    lp = rp = NULL;
    tp = np->next;

    while ( tp != NULL ) {
	bp = tp->next;
	if ( strcmp(np->name, tp->name) >= 0 ) {
	    tp->next = lp;
	    lp = tp;
	} else {
	    tp->next = rp;
	    rp = tp;
	}
	tp = bp;
    }

    lp = dir_sort(lp);
    rp = dir_sort(rp);

    if ( lp != NULL ) {
	tp = lp;
	while ( lp->next != NULL )
	    lp = lp->next;
	lp->next = np;
    } else
	tp = np;

    np->next = rp;

    return tp;
}
DIRPTR	*dir_get(char *dir)
{
    DIR    *dirp;
    struct dirent *dp;
    DIRPTR *tp, *top, *np;

    if ( (dirp = opendir(dir)) == NULL ) {
	printf("%s:can't open dir '%s'\n", prog, dir);
	return NULL;
    }

    readdir(dirp);	/* . */
    readdir(dirp);	/* .. */

    top = np = NULL;

    while ( (dp = readdir(dirp)) != NULL ) {
	if ( (tp = (DIRPTR *)malloc(sizeof(DIRPTR) + 
				    strlen(dp->d_name))) == NULL )
	    break;

	strcpy(tp->name, dp->d_name);

	if ( np == NULL )
	    top = tp;
	else
	    np->next = tp;

	np = tp;
	np->next = NULL;
    }

    closedir(dirp);

    if ( sort_flg )
	top = dir_sort(top);

    return top;
}

int	strklen(char *str)
{
    int len;

    for ( len = 0 ; *str != '\0' ; len++ ) {
	if ( str[0] == '\x8E' && str[1] != '\0' )
	    str += 2;
	else
	    str += 1;
    }
    return len;
}
char	*file_time(time_t t)
{
    static char tmp[32];
    struct tm *tm;

    tm = localtime(&t);
    sprintf(tmp, "%02d-%02d-%02d %02d:%02d",
	tm->tm_year % 100, tm->tm_mon + 1, tm->tm_mday,
	tm->tm_hour, tm->tm_min);
    return tmp;
}
int	file_type(int at)
{
    if ( ext_disp ) {
    	switch(at & S_IFMT) {
    	case S_IFDIR:  return '/';
    	case S_IFCHR:  return '<';
    	case S_IFBLK:  return '+';
    	case S_IFLNK:  return '@';
    	case S_IFIFO:  return '&';
    	case S_IFSOCK: return '=';
    	}
    	if ( (at & (S_IXUSR | S_IXGRP | S_IXOTH)) != 0 )
	    return '*';
    }
    return ' ';
}
char	*uidname(int uid)
{
    static char uname[32];
    static int last_uid = -1;
    struct passwd *pwent;

    if( uid == last_uid )
	return uname;

    *uname = 0;
    if ( (pwent = getpwuid(uid)) == NULL )
	sprintf(uname,"%d",uid);
    else
	strcpy(uname,pwent->pw_name);
    last_uid = uid;
    return uname;
}
void	file_print(char *nst, char *sep, char *file, struct stat *st)
{
    int     i, n;

    printf("%s%s%s%c ", nst, sep, file, file_type(st->st_mode));
    n = (uid_disp ? 41:50) -
	(strlen(nst) + strklen(file) + 2 + (*sep == '\0' ? 0:3));
    for ( i = n % 3 ; i > 0 ; i-- )
	putchar(' ');
    for ( i = n / 3 ; i > 0 ; i-- )
	printf("  -");
    if ( uid_disp )
	printf(" %-8s", uidname(st->st_uid));
    printf(" %9qd %s\n", st->st_size, file_time(st->st_mtime));
}
void	tree(char *dir)
{
    int     i, n;
    char    *p, *s;
    DIRPTR  *dp, *tp;
    struct stat st;

    for ( p = nest ; *p != '\0' ; p++ )
	;

    for ( s = cwd ; *s != '\0' ; s++ )
	;

    if ( (dp = dir_get(dir)) == NULL )
	return;

    *s = '/';
    while ( dp != NULL ) {

	strcpy(s + 1, dp->name);

	if ( sym_disp )
            stat(cwd, &st);
	else
            lstat(cwd, &st);

	file_print(nest, (dp->next == NULL ? "`- ":"+- "), dp->name, &st);

	if ( red_disp && (st.st_mode & S_IFMT) == S_IFDIR ) {
	    strcpy(p, (dp->next == NULL ? "   ":"|  "));
	    tree(cwd);
	    *p = '\0';
	}

	tp = dp->next;
	free(dp);
	dp = tp;
    }

    *s = '\0';
}
int	main_tree(char *dir)
{
    int     i, n;
    char    *p, *s;
    DIRPTR  *dp, *tp;
    struct stat st;

    strcpy(cwd, dir);
    strcpy(nest, "  ");

    if ( stat(dir, &st) ) {
	fprintf(stderr, "%s: Can't stat %s\n", prog, dir);
	return ERR;
    }

    file_print("", "", dir, &st);

    if ( (st.st_mode & S_IFMT) == S_IFDIR )
    	tree(dir);

    return FALSE;
}

#define	OPT_LONG

#ifdef	OPT_LONG

#include    "opt.c"

option	opt_tab[] = {
	{ 'L', 0, "no_sym_link" },
	{ 'F', 0, "no_file_type" },
	{ 'U', 0, "no_uid" },
	{ 'R', 0, "no_recursive" },
	{ 'S', 0, "no_sort" },
	{ 0, 0, NULL },
};

#else

extern	char	*optarg;
extern	int	optind;

#endif

void	usage()
{
    fprintf(stderr, "Usage: %s [-LFURS]\n", prog);
    fprintf(stderr, 
	"\t[-L] [--no_sym_link] [-F] [--no_file_type] [-U] [--no_uid]\n"\
	"\t[-R] [--no_recursive] [-S] [--no_sort]\n");
}
int	main(ac,av)
int	ac;
char	*av[];
{
    int    opt;
    int    mx = 1;

    prog = av[0];

    for ( ; ; ) {
#ifdef	OPT_LONG
	while ( (opt = getopt_long(ac, av, opt_tab)) != EOF ) {
#else
	while ( (opt = getopt(ac,av,"LFRSU")) != EOF ) {
#endif
	    switch(opt) {
	    case 'L':
		sym_disp = FALSE;
		break;
	    case 'F':
		ext_disp = FALSE;
		break;
	    case 'R':
		red_disp = FALSE;
		break;
	    case 'U':
		uid_disp = FALSE;
		break;
	    case 'S':
		sort_flg = FALSE;
		break;
	    default:
		usage();
		exit(1);
	        break;
	    }
	}
	if ( optind >= ac )
	    break;
	av[mx++] = av[optind++];
    }

    if ( mx > 1 ) {
	while ( --mx > 0 )
	    main_tree(*(++av));
    } else
	main_tree(".");

    return 0;
}

